@@ -2,13 +2,15 @@ from __future__ import absolute_import, division, print_function
 
 import os
 import sys
+import warnings
 from collections import OrderedDict
 
 import numpy as np
 from pandas import ExcelWriter, ExcelFile, HDFStore
 
-from .core import LArray, Axis, read_csv, read_hdf, df_aslarray, larray_equal, larray_nan_equal, get_axes
+from .core import LArray, Axis, read_csv, read_hdf, df_aslarray, larray_nan_equal, get_axes
 from .excel import open_excel
+from .utils import float_error_handler_factory, pickle
 
 
 def check_pattern(k, pattern):
@@ -79,19 +81,18 @@ class FileHandler(object):
 
         Returns
         -------
-        dict(str,LArray)
-            Dictionary containing names and arrays loaded from a file.
+        OrderedDict(str, LArray)
+            Dictionary containing the loaded arrays.
         """
         display = kwargs.pop('display', False)
         self._open_for_read()
-        res = {}
+        res = OrderedDict()
         if keys is None:
             keys = self.list()
         for key in keys:
             if display:
                 print("loading", key, "...", end=' ')
-            dest_key = key.strip('/')
-            res[dest_key] = self._read_array(key, *args, **kwargs)
+            res[key] = self._read_array(key, *args, **kwargs)
             if display:
                 print("done")
         self.close()
@@ -104,11 +105,10 @@ class FileHandler(object):
 
         Parameters
         ----------
-        key_values : dict of paris (str, LArray)
-            Dictionary containing arrays to dump.
+        key_values : list of (str, LArray) pairs
+            Name and data of arrays to dump.
         kwargs :
-            * display: a small message is displayed to tell when
-              an array is started to be dump and when it's done.
+            * display: whether or not to display when the dump of each array is started/done.
         """
         display = kwargs.pop('display', False)
         self._open_for_write()
@@ -133,13 +133,16 @@ class PandasHDFHandler(FileHandler):
         self.handle = HDFStore(self.fname)
 
     def list(self):
-        return self.handle.keys()
+        return [key.strip('/') for key in self.handle.keys()]
+
+    def _to_hdf_key(self, key):
+        return '/' + key
 
     def _read_array(self, key, *args, **kwargs):
-        return read_hdf(self.handle, key, *args, **kwargs)
+        return read_hdf(self.handle, self._to_hdf_key(key), *args, **kwargs)
 
     def _dump(self, key, value, *args, **kwargs):
-        value.to_hdf(self.handle, key, *args, **kwargs)
+        value.to_hdf(self.handle, self._to_hdf_key(key), *args, **kwargs)
 
     def close(self):
         self.handle.close()
@@ -201,41 +204,72 @@ class PandasCSVHandler(FileHandler):
         pass
 
     def _open_for_write(self):
-        try:
-            os.makedirs(self.fname)
-        except OSError:
-            if not os.path.isdir(self.fname):
-                raise
+        if self.fname is not None:
+            try:
+                os.makedirs(self.fname)
+            except OSError:
+                if not os.path.isdir(self.fname):
+                    raise ValueError("Path {} must represent a directory".format(self.fname))
 
     def list(self):
         # strip extension from files
-        # FIXME: only take .csv files
         # TODO: also support fname pattern, eg. "dump_*.csv" (using glob)
-        return [os.path.splitext(fname)[0] for fname in os.listdir(self.fname)]
+        if self.fname is not None:
+            return sorted([os.path.splitext(fname)[0] for fname in os.listdir(self.fname) if '.csv' in fname])
+        else:
+            return []
+
+    def _to_filepath(self, key):
+        if self.fname is not None:
+            return os.path.join(self.fname, '{}.csv'.format(key))
+        else:
+            return key
 
     def _read_array(self, key, *args, **kwargs):
-        fpath = os.path.join(self.fname, '{}.csv'.format(key))
-        return read_csv(fpath, *args, **kwargs)
+        return read_csv(self._to_filepath(key), *args, **kwargs)
 
     def _dump(self, key, value, *args, **kwargs):
-        value.to_csv(os.path.join(self.fname, '{}.csv'.format(key)), *args,
-                     **kwargs)
+        value.to_csv(self._to_filepath(key), *args, **kwargs)
 
     def close(self):
         pass
 
 
+class PickleHandler(FileHandler):
+    def _open_for_read(self):
+        with open(self.fname, 'rb') as f:
+            self.data = pickle.load(f)
+
+    def _open_for_write(self):
+        self.data = OrderedDict()
+
+    def list(self):
+        return self.data.keys()
+
+    def _read_array(self, key):
+        return self.data[key]
+
+    def _dump(self, key, value):
+        self.data[key] = value
+
+    def close(self):
+        with open(self.fname, 'wb') as f:
+            pickle.dump(self.data, f)
+
+
 handler_classes = {
+    'pickle': PickleHandler,
+    'pandas_csv': PandasCSVHandler,
     'pandas_hdf': PandasHDFHandler,
     'pandas_excel': PandasExcelHandler,
     'xlwings_excel': XLWingsHandler,
-    'pandas_csv': PandasCSVHandler
 }
 
 ext_default_engine = {
+    'csv': 'pandas_csv',
     'h5': 'pandas_hdf', 'hdf': 'pandas_hdf',
+    'pkl': 'pickle', 'pickle': 'pickle',
     'xls': 'xlwings_excel', 'xlsx': 'xlwings_excel',
-    'csv': 'pandas_csv'
 }
 
 
@@ -287,6 +321,9 @@ class Session(object):
         for k, v in kwargs.items():
             self[k] = v
 
+    def _ipython_key_completions_(self):
+        return list(self.keys())
+
     def __getitem__(self, key):
         if isinstance(key, int):
             return self._objects[self.keys()[key]]
@@ -330,23 +367,47 @@ class Session(object):
     def __setitem__(self, key, value):
         self._objects[key] = value
 
+    def __delitem__(self, key):
+        del self._objects[key]
+
     def __getattr__(self, key):
-        return self._objects[key]
+        if key in self._objects:
+            return self._objects[key]
+        else:
+            raise AttributeError("'{}' object has no attribute '{}'".format(self.__class__.__name__, key))
 
     def __setattr__(self, key, value):
         self._objects[key] = value
 
+    def __delattr__(self, key):
+        del self._objects[key]
+
+    def __dir__(self):
+        return list(set(dir(self.__class__)) | set(self.keys()))
+
+    # needed to make *un*pickling work (because otherwise, __getattr__ is called before ._objects exists, which leads to
+    # an infinite recursion)
+    def __getstate__(self):
+        return self.__dict__
+
+    def __setstate__(self, d):
+        # equivalent to self.__dict__ = d (we need this form because __setattr__ is overridden)
+        object.__setattr__(self, '__dict__', d)
+
     def load(self, fname, names=None, engine='auto', display=False, **kwargs):
         """
         Loads array objects from a file.
 
+        WARNING: never load a file using the pickle engine (.pkl or .pickle) from an untrusted source, as it can lead
+        to arbitrary code execution.
+
         Parameters
         ----------
         fname : str
             Path to the file.
         names : list of str, optional
-            List of arrays to load. Defaults to all valid objects present in
-            the file/directory.
+            List of arrays to load. If `fname` is None, list of paths to CSV files.
+            Defaults to all valid objects present in the file/directory.
         engine : str, optional
             Load using `engine`. Defaults to 'auto' (use default engine for
             the format guessed from the file extension).
@@ -356,17 +417,22 @@ class Session(object):
         """
         if display:
             print("opening", fname)
-        # TODO: support path + *.csv
+        if fname is None:
+            if all([os.path.splitext(name)[1] == '.csv' for name in names]):
+                engine = ext_default_engine['csv']
+            else:
+                raise ValueError("List of paths to only CSV files expected. Got {}".format(names))
         if engine == 'auto':
             _, ext = os.path.splitext(fname)
-            engine = ext_default_engine[ext.strip('.')]
+            ext = ext.strip('.') if '.' in ext else 'csv'
+            engine = ext_default_engine[ext]
         handler_cls = handler_classes[engine]
         handler = handler_cls(fname)
         arrays = handler.read_arrays(names, display=display, **kwargs)
         for k, v in arrays.items():
             self[k] = v
 
-    def dump(self, fname, names=None, engine='auto', display=False, **kwargs):
+    def save(self, fname, names=None, engine='auto', display=False, **kwargs):
         """
         Dumps all array objects from the current session to a file.
 
@@ -375,8 +441,8 @@ class Session(object):
         fname : str
             Path for the dump.
         names : list of str or None, optional
-            List of names of objects to dump. Defaults to all objects
-            present in the Session.
+            List of names of objects to dump. If `fname` is None, list of paths to CSV files.
+            Defaults to all objects present in the Session.
         engine : str, optional
             Dump using `engine`. Defaults to 'auto' (use default engine for
             the format guessed from the file extension).
@@ -386,18 +452,38 @@ class Session(object):
         """
         if engine == 'auto':
             _, ext = os.path.splitext(fname)
-            engine = ext_default_engine[ext.strip('.')]
+            ext = ext.strip('.') if '.' in ext else 'csv'
+            engine = ext_default_engine[ext]
         handler_cls = handler_classes[engine]
         handler = handler_cls(fname)
-        filtered = self.filter(kind=LArray)
-        # not using .items() so that arrays are sorted
-        arrays = [(k, filtered[k]) for k in filtered.names]
+        items = self.filter(kind=LArray).items()
         if names is not None:
             names_set = set(names)
-            arrays = [(k, v) for k, v in arrays if k in names_set]
-        handler.dump_arrays(arrays, display=display, **kwargs)
+            items = [(k, v) for k, v in items if k in names_set]
+        handler.dump_arrays(items, display=display, **kwargs)
 
-    def dump_hdf(self, fname, names=None, *args, **kwargs):
+    def to_pickle(self, fname, names=None, *args, **kwargs):
+        """
+        Dumps all array objects from the current session to a file using pickle.
+
+        WARNING: never load a pickle file (.pkl or .pickle) from an untrusted source, as it can lead to arbitrary code
+        execution.
+
+        Parameters
+        ----------
+        fname : str
+            Path for the dump.
+        names : list of str or None, optional
+            List of names of objects to dump. Defaults to all objects
+            present in the Session.
+        """
+        self.save(fname, names, ext_default_engine['pkl'], *args, **kwargs)
+
+    def dump(self, fname, names=None, engine='auto', display=False, **kwargs):
+        warnings.warn("Method dump is deprecated. Use method save instead.", DeprecationWarning, stacklevel=2)
+        self.save(fname, names, engine, display, **kwargs)
+
+    def to_hdf(self, fname, names=None, *args, **kwargs):
         """
         Dumps all array objects from the current session to an HDF file.
 
@@ -409,9 +495,13 @@ class Session(object):
             List of names of objects to dump. Defaults to all objects
             present in the Session.
         """
-        self.dump(fname, names, ext_default_engine['hdf'], *args, **kwargs)
+        self.save(fname, names, ext_default_engine['hdf'], *args, **kwargs)
 
-    def dump_excel(self, fname, names=None, *args, **kwargs):
+    def dump_hdf(self, fname, names=None, *args, **kwargs):
+        warnings.warn("Method dump_hdf is deprecated. Use method to_hdf instead.", DeprecationWarning, stacklevel=2)
+        self.to_hdf(fname, names, *args, **kwargs)
+
+    def to_excel(self, fname, names=None, *args, **kwargs):
         """
         Dumps all array objects from the current session to an Excel file.
 
@@ -423,9 +513,13 @@ class Session(object):
             List of names of objects to dump. Defaults to all objects
             present in the Session.
         """
-        self.dump(fname, names, ext_default_engine['xlsx'], *args, **kwargs)
+        self.save(fname, names, ext_default_engine['xlsx'], *args, **kwargs)
 
-    def dump_csv(self, fname, names=None, *args, **kwargs):
+    def dump_excel(self, fname, names=None, *args, **kwargs):
+        warnings.warn("Method dump_excel is deprecated. Use method to_excel instead.", DeprecationWarning, stacklevel=2)
+        self.to_excel(fname, names, *args, **kwargs)
+
+    def to_csv(self, fname, names=None, *args, **kwargs):
         """
         Dumps all array objects from the current session to a CSV file.
 
@@ -437,7 +531,11 @@ class Session(object):
             List of names of objects to dump. Defaults to all objects
             present in the Session.
         """
-        self.dump(fname, names, ext_default_engine['csv'], *args, **kwargs)
+        self.save(fname, names, ext_default_engine['csv'], *args, **kwargs)
+
+    def dump_csv(self, fname, names=None, *args, **kwargs):
+        warnings.warn("Method dump_csv is deprecated. Use method to_csv instead.", DeprecationWarning, stacklevel=2)
+        self.to_csv(fname, names, *args, **kwargs)
 
     def filter(self, pattern=None, kind=None):
         """
@@ -455,15 +553,12 @@ class Session(object):
         Session
             The filtered session.
         """
+        items = self._objects.items()
         if pattern is not None:
-            items = [(k, self._objects[k]) for k in self._objects.keys()
-                     if check_pattern(k, pattern)]
-        else:
-            items = self._objects.items()
+            items = [(k, v) for k, v in items if check_pattern(k, pattern)]
         if kind is not None:
-            return Session([(k, v) for k, v in items if isinstance(v, kind)])
-        else:
-            return Session(items)
+            items = [(k, v) for k, v in items if isinstance(v, kind)]
+        return Session(items)
 
     @property
     def names(self):
@@ -479,6 +574,7 @@ class Session(object):
     def copy(self):
         """Returns a copy of the session.
         """
+        # this actually *does* a copy of the internal mapping (the mapping is not reused-as is)
         return Session(self._objects)
 
     def keys(self):
@@ -505,11 +601,12 @@ class Session(object):
             self_keys = set(self.keys())
             all_keys = list(self.keys()) + [n for n in other.keys() if
                                             n not in self_keys]
-            res = []
-            for name in all_keys:
-                self_array = self.get(name, np.nan)
-                other_array = other.get(name, np.nan)
-                res.append((name, getattr(self_array, opfullname)(other_array)))
+            with np.errstate(call=_session_float_error_handler):
+                res = []
+                for name in all_keys:
+                    self_array = self.get(name, np.nan)
+                    other_array = other.get(name, np.nan)
+                    res.append((name, getattr(self_array, opfullname)(other_array)))
             return Session(res)
         opmethod.__name__ = opfullname
         return opmethod
@@ -526,7 +623,7 @@ class Session(object):
         all_keys = list(self.keys()) + [n for n in other.keys()
                                         if n not in self_keys]
         res = [larray_nan_equal(self.get(key), other.get(key)) for key in all_keys]
-        return LArray(res, [Axis('name', all_keys)])
+        return LArray(res, [Axis(all_keys, 'name')])
 
     def __ne__(self, other):
         return ~(self == other)
@@ -568,3 +665,6 @@ def local_arrays(depth=0):
     d = sys._getframe(depth + 1).f_locals
     return Session((k, d[k]) for k in sorted(d.keys())
                    if isinstance(d[k], LArray))
+
+
+_session_float_error_handler = float_error_handler_factory(4)
