@@ -1,12 +1,18 @@
 from __future__ import absolute_import, division, print_function
 
 from unittest import TestCase
-import unittest
 
+import pytest
 import numpy as np
 
 from larray import Session, Axis, LArray, ndrange, isnan, larray_equal
 from larray.tests.test_la import assert_array_nan_equal, abspath
+from larray.utils import pickle
+
+try:
+    import xlwings as xw
+except ImportError:
+    xw = None
 
 
 def equal(o1, o2):
@@ -20,17 +26,18 @@ def equal(o1, o2):
 
 class TestSession(TestCase):
     def setUp(self):
-        self.a = Axis('a', [])
-        self.b = Axis('b', [])
+        self.a = Axis([], 'a')
+        self.b = Axis([], 'b')
         self.c = 'c'
         self.d = {}
-        self.e = ndrange([('a0', 2), ('a1', 3)])
-        self.f = ndrange([('a0', 3), ('a1', 2)])
-        self.g = ndrange([('a0', 2), ('a1', 4)])
-        self.session = Session([('b', self.b), ('a', self.a),
-                                ('c', self.c), ('d', self.d),
-                                ('e', self.e), ('f', self.f),
-                                ('g', self.g)])
+        self.e = ndrange([(2, 'a0'), (3, 'a1')])
+        self.f = ndrange([(3, 'a0'), (2, 'a1')])
+        self.g = ndrange([(2, 'a0'), (4, 'a1')])
+        self.session = Session([
+            ('b', self.b), ('a', self.a),
+            ('c', self.c), ('d', self.d),
+            ('e', self.e), ('g', self.g), ('f', self.f),
+        ])
 
     def assertObjListEqual(self, got, expected):
         self.assertEqual(len(got), len(expected))
@@ -95,26 +102,26 @@ class TestSession(TestCase):
 
     def test_add(self):
         s = self.session
-        h = Axis('h', [])
+        h = Axis([], 'h')
         s.add(h, i='i')
         self.assertTrue(h.equals(s.h))
         self.assertEqual(s.i, 'i')
 
     def test_iter(self):
-        expected = [self.b, self.a, self.c, self.d, self.e, self.f, self.g]
+        expected = [self.b, self.a, self.c, self.d, self.e, self.g, self.f]
         self.assertObjListEqual(self.session, expected)
 
     def test_filter(self):
         s = self.session
         s.ax = 'ax'
         self.assertObjListEqual(s.filter(), [self.b, self.a, 'c', {},
-                                             self.e, self.f, self.g, 'ax'])
+                                             self.e, self.g, self.f, 'ax'])
         self.assertEqual(list(s.filter('a')), [self.a, 'ax'])
         self.assertEqual(list(s.filter('a', dict)), [])
         self.assertEqual(list(s.filter('a', str)), ['ax'])
         self.assertEqual(list(s.filter('a', Axis)), [self.a])
         self.assertEqual(list(s.filter(kind=Axis)), [self.b, self.a])
-        self.assertObjListEqual(s.filter(kind=LArray), [self.e, self.f, self.g])
+        self.assertObjListEqual(s.filter(kind=LArray), [self.e, self.g, self.f])
         self.assertEqual(list(s.filter(kind=dict)), [{}])
 
     def test_names(self):
@@ -128,32 +135,60 @@ class TestSession(TestCase):
     def test_h5_io(self):
         fpath = abspath('test_session.h5')
 
-        self.session.dump(fpath)
+        self.session.save(fpath)
         s = Session()
         s.load(fpath)
-        self.assertEqual(s.names, ['e', 'f', 'g'])
+        # HDF does *not* keep ordering (ie, keys are always sorted)
+        self.assertEqual(list(s.keys()), ['e', 'f', 'g'])
 
         s = Session()
         s.load(fpath, ['e', 'f'])
-        self.assertEqual(s.names, ['e', 'f'])
-
-    def test_xlsx_io(self):
-        self.session.dump(abspath('test_session.xlsx'), engine='pandas_excel')
-        self.session.dump(abspath('test_session_ef.xlsx'), ['e', 'f'], engine='pandas_excel')
-        # dump_excel uses default engine (xlwings) which is not available on
-        # travis
-        # self.session.dump_excel('test_session2.xlsx')
+        self.assertEqual(list(s.keys()), ['e', 'f'])
 
+    def test_xlsx_pandas_io(self):
+        fpath = abspath('test_session.xlsx')
+        self.session.save(fpath, engine='pandas_excel')
         s = Session()
-        s.load(abspath('test_session_ef.xlsx'), engine='pandas_excel')
-        self.assertEqual(s.names, ['e', 'f'])
+        s.load(fpath, engine='pandas_excel')
+        self.assertEqual(list(s.keys()), ['e', 'g', 'f'])
+
+        fpath = abspath('test_session_ef.xlsx')
+        self.session.save(fpath, ['e', 'f'], engine='pandas_excel')
+        s = Session()
+        s.load(fpath, engine='pandas_excel')
+        self.assertEqual(list(s.keys()), ['e', 'f'])
+
+    @pytest.mark.skipif(xw is None, reason="xlwings is not available")
+    def test_xlsx_xlwings_io(self):
+        fpath = abspath('test_session_xw.xlsx')
+        self.session.save(fpath, engine='xlwings_excel')
+        s = Session()
+        s.load(fpath, engine='xlwings_excel')
+        # ordering is only kept if the file did not exist previously (otherwise the ordering is left intact)
+        self.assertEqual(list(s.keys()), ['e', 'g', 'f'])
+
+        fpath = abspath('test_session_ef_xw.xlsx')
+        self.session.save(fpath, ['e', 'f'], engine='xlwings_excel')
+        s = Session()
+        s.load(fpath, engine='xlwings_excel')
+        self.assertEqual(list(s.keys()), ['e', 'f'])
 
     def test_csv_io(self):
-        self.session.dump_csv(abspath('test_session_csv'))
+        fpath = abspath('test_session_csv')
+        self.session.to_csv(fpath)
 
         s = Session()
-        s.load(abspath('test_session_csv'), engine='pandas_csv')
-        self.assertEqual(s.names, ['e', 'f', 'g'])
+        s.load(fpath, engine='pandas_csv')
+        # CSV cannot keep ordering (so we always sort keys)
+        self.assertEqual(list(s.keys()), ['e', 'f', 'g'])
+
+    def test_pickle_io(self):
+        fpath = abspath('test_session.pkl')
+
+        self.session.save(fpath)
+        s = Session()
+        s.load(fpath, engine='pickle')
+        self.assertEqual(list(s.keys()), ['e', 'g', 'f'])
 
     def test_eq(self):
         sess = self.session.filter(kind=LArray)
@@ -164,16 +199,16 @@ class TestSession(TestCase):
         res = sess == other
         self.assertEqual(res.ndim, 1)
         self.assertEqual(res.axes.names, ['name'])
-        self.assertTrue(np.array_equal(res.axes.labels[0], ['e', 'f', 'g']))
-        self.assertEqual(list(res), [True, True, False])
+        self.assertTrue(np.array_equal(res.axes.labels[0], ['e', 'g', 'f']))
+        self.assertEqual(list(res), [True, False, True])
 
         e2 = self.e.copy()
         e2.i[1, 1] = 42
         other = Session({'e': e2, 'f': self.f})
         res = sess == other
         self.assertEqual(res.axes.names, ['name'])
-        self.assertTrue(np.array_equal(res.axes.labels[0], ['e', 'f', 'g']))
-        self.assertEqual(list(res), [False, True, False])
+        self.assertTrue(np.array_equal(res.axes.labels[0], ['e', 'g', 'f']))
+        self.assertEqual(list(res), [False, False, True])
 
     def test_ne(self):
         sess = self.session.filter(kind=LArray)
@@ -183,16 +218,16 @@ class TestSession(TestCase):
         other = Session({'e': self.e, 'f': self.f})
         res = sess != other
         self.assertEqual(res.axes.names, ['name'])
-        self.assertTrue(np.array_equal(res.axes.labels[0], ['e', 'f', 'g']))
-        self.assertEqual(list(res), [False, False, True])
+        self.assertTrue(np.array_equal(res.axes.labels[0], ['e', 'g', 'f']))
+        self.assertEqual(list(res), [False, True, False])
 
         e2 = self.e.copy()
         e2.i[1, 1] = 42
         other = Session({'e': e2, 'f': self.f})
         res = sess != other
         self.assertEqual(res.axes.names, ['name'])
-        self.assertTrue(np.array_equal(res.axes.labels[0], ['e', 'f', 'g']))
-        self.assertEqual(list(res), [True, False, True])
+        self.assertTrue(np.array_equal(res.axes.labels[0], ['e', 'g', 'f']))
+        self.assertEqual(list(res), [True, True, False])
 
     def test_sub(self):
         sess = self.session.filter(kind=LArray)
@@ -205,9 +240,15 @@ class TestSession(TestCase):
     def test_div(self):
         sess = self.session.filter(kind=LArray)
         other = Session({'e': self.e - 1, 'f': self.f + 1})
-        res = sess / other
 
-        flat_e = np.arange(6) / np.arange(-1, 5)
+        with pytest.warns(RuntimeWarning) as caught_warnings:
+            res = sess / other
+        assert len(caught_warnings) == 1
+        assert caught_warnings[0].message.args[0] == "divide by zero encountered during operation"
+        assert caught_warnings[0].filename == __file__
+
+        with np.errstate(divide='ignore', invalid='ignore'):
+            flat_e = np.arange(6) / np.arange(-1, 5)
         assert_array_nan_equal(res['e'], flat_e.reshape(2, 3))
 
         flat_f = np.arange(6) / np.arange(1, 7)
@@ -218,11 +259,19 @@ class TestSession(TestCase):
         sess = self.session.filter(kind=LArray)
         self.assertEqual(sess.summary(),
                          "e: a0*, a1*\n    \n\n"
-                         "f: a0*, a1*\n    \n\n"
-                         "g: a0*, a1*\n    \n")
+                         "g: a0*, a1*\n    \n\n"
+                         "f: a0*, a1*\n    \n")
+
+    def test_pickle_roundtrip(self):
+        original = self.session
+        s = pickle.dumps(original)
+        res = pickle.loads(s)
+        self.assertTrue(all(res == original))
 
 
 if __name__ == "__main__":
     # import doctest
+    # import unittest
     # doctest.testmod(larray.core)
-    unittest.main()
+    # unittest.main()
+    pytest.main()
